#include "fraise.h"
void showFeatures(CvPoint2D32f *feats, int featcnt, IplImage **img){
  int icnt;
  for (icnt=0; icnt<featcnt; icnt++) {
    cvCircle(*img, cvPointFrom32f(feats[icnt]), 2, CV_RGB(255, 0, 0),
             0.5, 8, 0);
  }
}
void displayElementsWI (const CvMat *x, const char *marker) {
  int icnt, jcnt;
  fprintf(stderr, "%dx%d matrix: %s\n", x->rows, x->cols, marker);
  for (icnt =0; icnt < x->rows; icnt++){
    for (jcnt =0; jcnt < x->cols; jcnt++)
      fprintf(stderr, "%2.5f ", cvmGet(x, icnt, jcnt));
    fprintf(stderr, "\n");
  }
}
void printMat2file (const CvMat *x, const char *fname) {
  FILE *fp;
  int icnt, jcnt;

  fp = fopen(fname, "w");
  if (fp!=NULL) {
    for (icnt =0; icnt < x->rows; icnt++){
      for (jcnt =0; jcnt < x->cols; jcnt++)
        fprintf(fp, "%5.9f ",  (double)cvmGet(x, icnt, jcnt));
      fprintf(fp, "\n");
    }
    fclose(fp);
  }
}
void getFilelist(vector<string>& list, string dir, string extension) {
#ifdef WIN32
  WIN32_FIND_DATA findData;
  HANDLE fileHandle;
  int flag = 1;

  std::string search ("*.");
  if (dir == "") dir = ".";
  dir += "/";
  search = dir + search + extension;

  fileHandle = FindFirstFile(search.c_str(), &findData);
  if (fileHandle == INVALID_HANDLE_VALUE) return;
  while (flag) {
    list.push_back(findData.cFileName);
    flag = FindNextFile(fileHandle, &findData);
  }
  FindClose(fileHandle);
#else
  std::string search(".");
  search += extension;
  DIR* directory = opendir(dir.c_str());
  struct dirent* entry;
  std::string temp;

  while ((entry = readdir(directory)) != NULL) {
    temp = entry->d_name;
    if (temp.find(search, 0) != string::npos)
      list.push_back(temp);
  }
  closedir(directory);
#endif
  std::sort(list.begin(), list.end());
}

IplImage* loadReferenceImage(const char *dir, const char *file, const int conv2gray, const double rsz){
  IplImage *img =0, *imgrsz=0;
  char *filepath;

  filepath = (char *)malloc(MAX_PATH_LEN * sizeof(char ));
  bzero (filepath, MAX_PATH_LEN);
  sprintf (filepath, "%s%s%s", dir, _PATH_SEP, file); 
  
  if (conv2gray)
    img = (IplImage *)cvLoadImage(filepath, 0);
  else
    img = (IplImage *)cvLoadImage(filepath, -1);

  if(!img){
    printf("Could not load image file: %s\n", filepath);
    free(filepath);
    abort();
  }
  free(filepath);

  if (rsz !=1) {
    imgrsz = cvCreateImage(cvSize((int)ceil(img->width*rsz), (int)ceil(img->height*rsz)), img->depth, img->nChannels);
    cvResize(img, imgrsz, CV_INTER_LINEAR);
    cvReleaseImage(&img);
    return imgrsz;
  } else
    return img;
}
void _check_directories(const char *outputdir){
#ifdef WIN32
  ::CreateDirectory(outputdir, NULL);
#else
  if (opendir(outputdir) == NULL){
    fprintf(stderr, "ERR:Directory %s is not present\n", outputdir);
    if((mkdir(outputdir, 00777))== -1) {
      fprintf(stderr,"ERR:Error in creating dir\n");
      abort();
    }
    fprintf(stderr, "INF:Directory %s is created successfully\n", outputdir);
  }
#endif
}
void createMosaic(vector<IplImage *>& images, CvMat **H, const char *fname) {

  CvMat *corners = cvCreateMat(3, 4*images.size(), CV_32FC1);
  CvMat *loc = cvCreateMat(3, 1, CV_32FC1);
  for(int i=0; i < images.size(); ++i) {
    CvMat* tH = H[i];
    //0, 0
    cvmSet (loc, 0, 0, (double)0);
    cvmSet (loc, 1, 0, (double)0);
    cvmSet (loc, 2, 0, (double)1);
    cvGEMM (tH, loc, 1.0, NULL, 0.0, loc);
    cvmSet (corners, 0, 4*i, (double)cvmGet(loc, 0, 0));
    cvmSet (corners, 1, 4*i, (double)cvmGet(loc, 1, 0));
    cvmSet (corners, 2, 4*i, (double)cvmGet(loc, 2, 0));
    //width, 0
    cvmSet (loc, 0, 0, (double)images[i]->width);
    cvmSet (loc, 1, 0, (double)0);
    cvmSet (loc, 2, 0, (double)1);
    cvGEMM (tH, loc, 1.0, NULL, 0.0, loc);
    cvmSet (corners, 0, 4*i+1, (double)cvmGet(loc, 0, 0));
    cvmSet (corners, 1, 4*i+1, (double)cvmGet(loc, 1, 0));
    cvmSet (corners, 2, 4*i+1, (double)cvmGet(loc, 2, 0));
    //width, height
    cvmSet (loc, 0, 0, (double)images[i]->width);
    cvmSet (loc, 1, 0, (double)images[i]->height);
    cvmSet (loc, 2, 0, (double)1);
    cvGEMM (tH, loc, 1.0, NULL, 0.0, loc);
    cvmSet (corners, 0, 4*i+2, (double)cvmGet(loc, 0, 0));
    cvmSet (corners, 1, 4*i+2, (double)cvmGet(loc, 1, 0));
    cvmSet (corners, 2, 4*i+2, (double)cvmGet(loc, 2, 0));
    //0, height
    cvmSet (loc, 0, 0, (double)0);
    cvmSet (loc, 1, 0, (double)images[i]->height);
    cvmSet (loc, 2, 0, (double)1);
    cvGEMM (tH, loc, 1.0, NULL, 0.0, loc);
    cvmSet (corners, 0, 4*i+3, (double)cvmGet(loc, 0, 0));
    cvmSet (corners, 1, 4*i+3, (double)cvmGet(loc, 1, 0));
    cvmSet (corners, 2, 4*i+3, (double)cvmGet(loc, 2, 0));
  }
  //you should divide by third coordinate if projective
  double minX = DBL_MAX;
  double maxX = -DBL_MAX;
  double minY = DBL_MAX;
  double maxY = -DBL_MAX;
  for(int j = 0; j<images.size()*4; ++j) {
    if(cvmGet(corners, 0, j) < minX)
      minX = cvmGet(corners, 0, j);
    if(cvmGet(corners, 0, j) > maxX)
      maxX = cvmGet(corners, 0, j);
    if(cvmGet(corners, 1, j) < minY)
      minY = cvmGet(corners, 1, j);
    if(cvmGet(corners, 1, j) > maxY)
      maxY = cvmGet(corners, 1, j);
  }
  int outWidth = (int) (maxX - minX + 1);
  int outHeight = (int) (maxY - minY + 1);
  fprintf (stdout, "Resolution :%d x %d\n", outWidth, outHeight);
  
  CvMat *fit = cvCreateMat(3, 3, CV_32FC1);
  cvSetIdentity(fit);
  cvSet2D(fit, 0, 2, cvScalar(-minX));
  cvSet2D(fit, 1, 2, cvScalar(-minY));
  for(unsigned int i = 0; i < images.size(); ++i) {
    cvGEMM(fit, H[i], 1.0, NULL, 0.0, H[i]);
  }
  cvReleaseMat(&fit);
  
  IplImage *mosaic = cvCreateImage(cvSize(outWidth, outHeight), images[0]->depth, images[0]->nChannels);
  float *mosaicPtr = (float *) mosaic->imageData;
  cvSetZero(mosaic);
  CvMat *inv = cvCreateMat(3, 3, CV_32FC1);
  for(unsigned int k = 0; k < images.size(); ++k) {
    float *imgPtr = (float *) images[k]->imageData;
    fprintf (stdout, "Using image %d\n", k);
    cvInvert(H[k], inv);
    for(int i = 0; i < outHeight; ++i) {
      for(int j = 0; j < outWidth; ++j) {
        loc->data.fl[0] = (float) j;
        loc->data.fl[1] = (float) i;
        loc->data.fl[2] = 1;
        cvGEMM(inv, loc, 1.0, NULL, 0.0, loc);

        float x = loc->data.fl[0];
        float y = loc->data.fl[1];
        //if projective, divide by loc[2]

        int xF = (int) floorf(x);
        int xC = (int) ceilf(x);
        int yF = (int) floorf(y);
        int yC = (int) ceilf(y);
        if(xF < 0 || xF >= images[k]->width ||
             yF < 0 || yF >= images[k]->height ||
           xC < 0 || xC >= images[k]->width ||
             yC < 0 || yC >= images[k]->height)
          continue;
  
        //this assumes grayscale mosaic
        float val;
        //x and y integers
        if(x - xF < 0.000000001f && y - yF < 0.000000001f){
          val = imgPtr[yF*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0];
        }
        //x integer
        else if(x - xF < 0.000000001f) {
          val = (yC - y)*imgPtr[yF*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0] +
                (y - yF)*imgPtr[yC*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0];
        }
        //y integer
        else if(y - yF < 0.000000001f) {
          val = (xC - x)*imgPtr[yF*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0] +
                (x - xF)*imgPtr[yF*images[k]->width*images[k]->nChannels + xC*images[k]->nChannels + 0];
        }
        else {
          val = (xC - x)*(yC - y)*imgPtr[yF*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0] +
                (x - xF)*(yC - y)*imgPtr[yF*images[k]->width*images[k]->nChannels + xC*images[k]->nChannels + 0] +
              (xC - x)*(y - yF)*imgPtr[yC*images[k]->width*images[k]->nChannels + xF*images[k]->nChannels + 0] +
              (x - xF)*(y - yF)*imgPtr[yC*images[k]->width*images[k]->nChannels + xC*images[k]->nChannels + 0];
        }
        mosaicPtr[i*mosaic->width*mosaic->nChannels + j*mosaic->nChannels + 0] = val;
      }
    }
  }
  cvSaveImage (fname, mosaic, NULL);
  cvReleaseMat(&inv);
  cvReleaseMat(&corners);
  cvReleaseMat(&loc);
}
#ifdef HAVE_VIDEO
void generateVideo (vector<IplImage *>& frames, CvMat **H, const char *outfile){
  int _frame_count =0;
  CvVideoWriter *writer = NULL;
  CvSize frameSize;
  IplImage *res = NULL, *mask = NULL, *img = NULL; 
  double result;
	
  frameSize  = cvGetSize(frames[0]);
  /* Initializing a video writer object  with the input */	
  writer = cvCreateVideoWriter (outfile, CV_FOURCC('D','I','V','X'), 30, 
             frameSize, 0);
  	
  img = cvCreateImage (frameSize, frames[0]->depth, frames[0]->nChannels);
  /*fprintf (stdout, "frames[0]->depth %d: IPL_DEPTH_8U %d, \
             frames[0]->nChannels %d\n", 
		    frames[0]->depth, IPL_DEPTH_8U, frames[0]->nChannels);*/

  /* for storing the mask to specify the matrix elements to discard */	
  mask = cvCloneImage (img);
  cvSet (mask, cvRealScalar(255), NULL);
  res = cvCloneImage (img);
  while (_frame_count < frames.size()) {
    /* Warp the image first */
    cvWarpPerspective (frames[_frame_count], img, H[_frame_count],
      CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS+CV_WARP_INVERSE_MAP, cvScalarAll(0));
	/* Generating the mask for discarding the outlier pixels */
    cvWarpPerspective (mask, res, H[_frame_count],
      CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS+CV_WARP_INVERSE_MAP, cvScalarAll(0));
	  
	//cvShowImage ("mask", result);
    /* Qunatitative output - perform frame differencing between warped and 
			 original image, output result */		
	result = cvNorm(frames[_frame_count], img, CV_L2, res);
	
    fprintf (stdout, "||SRC - DST||_2 b/w frames %04d/%04d: %5.6f\n", 
		    _frame_count, _frame_count+1, result);
	
	cvWriteFrame (writer, img);
	_frame_count++;
  }
  /* clean up */	 
  cvReleaseImage (&img);
  cvReleaseImage (&res);
  cvReleaseImage (&mask);
  cvReleaseVideoWriter (&writer);
}
#endif

//Accept a vector of IplImages and an array of Homography matrices
void saveAlignedFrames(vector<IplImage *>& frames, CvMat **H, const char *outdir) {
  char *pref = (char *)malloc (MAX_PATH_LEN * sizeof(char));
  char *fname = (char *)malloc (MAX_PATH_LEN * sizeof(char));
  int _allocated = 0, _frame_count =0;

  while (_frame_count < frames.size()) {
    
    //Saving Aligned frames
    bzero(fname, MAX_PATH_LEN);
    sprintf(fname, "%04d%s", _frame_count, ".jpg");
    bzero(pref, MAX_PATH_LEN);
    sprintf(pref, "%s%svf_aligned", outdir, _PATH_SEP);
    saveResult(frames[_frame_count], H[_frame_count], fname, pref);
    _frame_count++;
  }
  // Cleaning up memory used 
  free (pref);
  free (fname);
}
/* Saving qualitative as well as qunatitative output for further analysis */
void saveResult(IplImage *imgsrc, CvMat *H, const char *origfname, const char *prefix) {
  char *fname;
  IplImage *img, *mask, *result;
	double res;
  static int frame = 0;
  if (H) {
    img = cvCreateImage (cvGetSize(imgsrc), imgsrc->depth, imgsrc->nChannels);
		/* for storing the mask : specify the matrix elements to discard */
		/* display images */
		//cvNamedWindow ("warped", CV_WINDOW_AUTOSIZE);
		//cvNamedWindow ("mask", CV_WINDOW_AUTOSIZE);
		
		mask = cvCloneImage (img);	
		result = cvCloneImage (img);
		cvSet (mask, cvRealScalar(255), NULL);
			
    cvWarpPerspective (imgsrc, img, H,
            CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS+CV_WARP_INVERSE_MAP, cvScalarAll(0));
		cvWarpPerspective (mask, result, H,
            CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS+CV_WARP_INVERSE_MAP, cvScalarAll(0));
		//cvShowImage ("mask", result);
		/* Qunatitative output - perform frame differencing between warped and 
			 original image, output result */		
	  res = cvNorm(imgsrc, img, CV_L2, result);
    fprintf (stdout, "||SRC - DST||_2 b/w frames %04d/%04d: %5.6f\n", 
		    frame, frame+1, res);
		frame++;
    fname = (char *) malloc (MAX_PATH_LEN* sizeof(char));
    sprintf (fname, "%s_%s", prefix, origfname);
    cvSaveImage (fname, img, NULL);
			
		//cvShowImage ("warped", img);
		//cvWaitKey (0);
		//cvDestroyAllWindows();
    free (fname);
  }
  cvReleaseImage (&img);
	cvReleaseImage (&mask);
	cvReleaseImage (&result);	
}
/* Argument 1 and 2 are only used if input is a video file
 assumes vH is a pre-allocated array of nframes CvMat matrices */
#ifdef HAVE_VIDEO
void outputRegParams(CvSeq *xseq, CvMat **vH, int nframes, const char *fname) {
  FILE *fp;
  int i,j, _frame_count;
  int _elem_count;
  CvMat *h;

  fp = fopen (fname, "w");
  if (fp) {
    _elem_count = 0;
    h  = cvCreateMat (3, 3, CV_32FC1);
    for (_frame_count=0; _frame_count <nframes; _frame_count++) {
      for (i =0; i <9; i++)
        cvmSet(h, i/3, i%3, *((double *)cvGetSeqElem (xseq, _elem_count++)));
      // save matrix elements in open file
      vH[_frame_count] = cvCloneMat(h);
      for (i =0; i < h->rows; i++){
        for (j =0; j < h->cols; j++)
          fprintf(fp, "%5.9f ",  (double)cvmGet(h, i, j));
        fprintf(fp, "\n");
      }
      fprintf(fp, "\n");
    }
    //Clean up
    cvReleaseMat (&h);
    fclose (fp);
  } else {
    fprintf (stderr, "ERR:Could not open %s for writing.\n", fname);
  }
}
#else
void outputRegParams(CvMat **xmat, int nframes, const char *fname) {
  FILE *fp;
  int i, j, _frame_count;

  fp = fopen (fname, "w");
  if (fp) {
    for (_frame_count=0; _frame_count <nframes; _frame_count++) {
      for (i =0; i < xmat[_frame_count]->rows; i++) {
        for (j =0; j< xmat[_frame_count]->cols; j++)
          fprintf(fp, "%5.9f ",  (double)cvmGet(xmat[_frame_count], i, j));
        fprintf(fp, "\n");
      }
      fprintf(fp, "\n");
    }
    fclose (fp);
  } else {
    fprintf (stderr, "ERR:Could not open %s for writing.\n", fname);
  }
}
#endif
